/*
    This file redefines several keys in the linux keymap,
    see kmap_worries.txt for more details.
*/


#include <linux/kd.h>
#include <linux/keyboard.h>

#define SHIFT_PRESSED       1
#define ALTGR_PRESSED       2
#define CTRL_PRESSED        4
#define SHIFTCTRL_PRESSED   5
#define ALT_PRESSED         8

//////////////////////////////////////////////////////////


int revise_keys()
{
    int
        fd=fileno(stdin),
        i=0;

    //  Be sure alt-f pulls down the File menu.
    //  This does not need superuser privileges
    //  and is very often already in force.
    ioctl(fd,KDSKBMETA,K_ESCPREFIX);

    //  be sure NumLock is on, ubuntu has other ideas
    //  K_NUMLOCK 0x02      from kd.h
    //  KDSKBLED  0x4B65    set led flags (not lights)
    //  KDSETLED  0x4B32    set led state (lights, not flags)
    ioctl(0, KDSKBLED, K_NUMLOCK);
    ioctl(0, KDSETLED, K_NUMLOCK);

    struct kbentry ke;

    int kmap[110];
    // Index is keycode, value is action code.
    // Array only sparsely used, redundance is convenience.
    // Backspace (14) is the lowest key we are going to touch.

    ke.kb_table=0;
    // We are reading the plain keys, no modifier.

    for (i=14;i<110;i++)
    {
        ke.kb_index=i;

        if (ioctl(fd, KDGKBENT, &ke))
        {
            close(fd);
            printf("\n\t*UNKNOWN READING ERROR* \n");
            return -i;
        }

        // When the first two entries in structure ke are set,
        // the third entry is filled in by the kernel.
        // We save it in our kmap[] array.

        kmap[i]=ke.kb_value;

        // This loop does not change anything, we are reading.
    }

    //  kmap[] has now the action values for the keys we target,
    //  essentially function keys, grey keys, arrows.
    //  we want to set modifier action values
    //  equal to the plain action values.
    //  this way, we get rid of fossilized keymap crap
    //  and can still identify the keys by reading modifiers.
    //  all despite the dire warnings in manpage
    //  console_ioctl which has been threatening
    //  fire and punishement since 1995.

    //  changes required because of significant differences
    //  e.g. between ubuntu and opensuse
    //  however, changes to keymap only allowed to superuser
    //  if we fail, we exit and display a message

    //  procedure:
    //  we try the change and we may fail for some reason,
    //  essentially for lacking permissions.
    //  however, before returning and complaining we test.
    //  it may well be that the intended change is already
    //  implemented and we do not care why, we just accept it.
    //  (some of the changes are indeed superfluous, they are
    //  there to make this code easier to read.)

    //  and chop your fingers off before condensing
    //  repetious code and making it unreadable

    // remember kmap[i] is now actioncode for plain key 'i'

    {
        // backspace
        i = 14;
        ke.kb_index=i;
        ke.kb_value=kmap[i];

        ke.kb_table=SHIFT_PRESSED;
        if (ioctl(fd,KDSKBENT,&ke))
        {
            //  actioncode not set: read and test it
            ioctl(fd, KDGKBENT, &ke);
            if(ke.kb_value != kmap[i])
            {
                close(fd);
                return -i;
            }
        }

        ke.kb_table=ALTGR_PRESSED;
        if (ioctl(fd,KDSKBENT,&ke))
        {
            //  actioncode not set: read and test it
            ioctl(fd, KDGKBENT, &ke);
            if(ke.kb_value != kmap[i])
            {
                close(fd);
                return -i;
            }
        }

        ke.kb_table=CTRL_PRESSED;
        if (ioctl(fd,KDSKBENT,&ke))
        {
            //  actioncode not set: read and test it
            ioctl(fd, KDGKBENT, &ke);
            if(ke.kb_value != kmap[i])
            {
                close(fd);
                return -i;
            }
        }

        ke.kb_table=SHIFTCTRL_PRESSED;
        if (ioctl(fd,KDSKBENT,&ke))
        {
            //  actioncode not set: read and test it
            ioctl(fd, KDGKBENT, &ke);
            if(ke.kb_value != kmap[i])
            {
                close(fd);
                return -i;
            }
        }

        ke.kb_table=ALT_PRESSED;
        if (ioctl(fd,KDSKBENT,&ke))
        {
            //  actioncode not set: read and test it
            ioctl(fd, KDGKBENT, &ke);
            if(ke.kb_value != kmap[i])
            {
                close(fd);
                return -i;
            }
        }
    }

    {
        // tab
        i = 15;
        ke.kb_index=i;
        ke.kb_value=kmap[i];

        ke.kb_table=SHIFT_PRESSED;
        if (ioctl(fd,KDSKBENT,&ke))
        {
            //  actioncode not set: read and test it
            ioctl(fd, KDGKBENT, &ke);
            if(ke.kb_value != kmap[i])
            {
                close(fd);
                return -i;
            }
        }

        ke.kb_table=ALTGR_PRESSED;
        if (ioctl(fd,KDSKBENT,&ke))
        {
            //  actioncode not set: read and test it
            ioctl(fd, KDGKBENT, &ke);
            if(ke.kb_value != kmap[i])
            {
                close(fd);
                return -i;
            }
        }

        ke.kb_table=CTRL_PRESSED;
        if (ioctl(fd,KDSKBENT,&ke))
        {
            //  actioncode not set: read and test it
            ioctl(fd, KDGKBENT, &ke);
            if(ke.kb_value != kmap[i])
            {
                close(fd);
                return -i;
            }
        }

        ke.kb_table=SHIFTCTRL_PRESSED;
        if (ioctl(fd,KDSKBENT,&ke))
        {
            //  actioncode not set: read and test it
            ioctl(fd, KDGKBENT, &ke);
            if(ke.kb_value != kmap[i])
            {
                close(fd);
                return -i;
            }
        }

        ke.kb_table=ALT_PRESSED;
        if (ioctl(fd,KDSKBENT,&ke))
        {
            //  actioncode not set: read and test it
            ioctl(fd, KDGKBENT, &ke);
            if(ke.kb_value != kmap[i])
            {
                close(fd);
                return -i;
            }
        }
    }

    for (i=59;i<69;i++)
    {
        //  F1, F2,.., F10

        ke.kb_index=i;
        ke.kb_value=kmap[i];

        ke.kb_table=SHIFT_PRESSED;
        if (ioctl(fd,KDSKBENT,&ke))
        {
            //  actioncode not set: read and test it
            ioctl(fd, KDGKBENT, &ke);
            if(ke.kb_value != kmap[i])
            {
                close(fd);
                return -i;
            }
        }

        ke.kb_table=ALTGR_PRESSED;
        if (ioctl(fd,KDSKBENT,&ke))
        {
            //  actioncode not set: read and test it
            ioctl(fd, KDGKBENT, &ke);
            if(ke.kb_value != kmap[i])
            {
                close(fd);
                return -i;
            }
        }

        ke.kb_table=CTRL_PRESSED;
        if (ioctl(fd,KDSKBENT,&ke))
        {
            //  actioncode not set: read and test it
            ioctl(fd, KDGKBENT, &ke);
            if(ke.kb_value != kmap[i])
            {
                close(fd);
                return -i;
            }
        }

        ke.kb_table=SHIFTCTRL_PRESSED;
        if (ioctl(fd,KDSKBENT,&ke))
        {
            //  actioncode not set: read and test it
            ioctl(fd, KDGKBENT, &ke);
            if(ke.kb_value != kmap[i])
            {
                close(fd);
                return -i;
            }
        }

        ke.kb_table=ALT_PRESSED;
        if (ioctl(fd,KDSKBENT,&ke))
        {
            //  actioncode not set: read and test it
            ioctl(fd, KDGKBENT, &ke);
            if(ke.kb_value != kmap[i])
            {
                close(fd);
                return -i;
            }
        }
    }

    for (i=87;i<89;i++)
    {
        // F11, F12

        ke.kb_index=i;
        ke.kb_value=kmap[i];

        ke.kb_table=SHIFT_PRESSED;
        if (ioctl(fd,KDSKBENT,&ke))
        {
            //  actioncode not set: read and test it
            ioctl(fd, KDGKBENT, &ke);
            if(ke.kb_value != kmap[i])
            {
                close(fd);
                return -i;
            }
        }

        ke.kb_table=ALTGR_PRESSED;
        if (ioctl(fd,KDSKBENT,&ke))
        {
            //  actioncode not set: read and test it
            ioctl(fd, KDGKBENT, &ke);
            if(ke.kb_value != kmap[i])
            {
                close(fd);
                return -i;
            }
        }

        ke.kb_table=CTRL_PRESSED;
        if (ioctl(fd,KDSKBENT,&ke))
        {
            //  actioncode not set: read and test it
            ioctl(fd, KDGKBENT, &ke);
            if(ke.kb_value != kmap[i])
            {
                close(fd);
                return -i;
            }
        }

        ke.kb_table=SHIFTCTRL_PRESSED;
        if (ioctl(fd,KDSKBENT,&ke))
        {
            //  actioncode not set: read and test it
            ioctl(fd, KDGKBENT, &ke);
            if(ke.kb_value != kmap[i])
            {
                close(fd);
                return -i;
            }
        }

        ke.kb_table=ALT_PRESSED;
        if (ioctl(fd,KDSKBENT,&ke))
        {
            //  actioncode not set: read and test it
            ioctl(fd, KDGKBENT, &ke);
            if(ke.kb_value != kmap[i])
            {
                close(fd);
                return -i;
            }
        }
    }

    for (i=103;i<110;i++)
    {
        //  arrows, PageUp, PageDown
        //  most settings here are suplerfluous
        //  but so the code is easier to type and read

        ke.kb_index=i;
        ke.kb_value=kmap[i];

        ke.kb_table=SHIFT_PRESSED;
        if (ioctl(fd,KDSKBENT,&ke))
        {
            //  actioncode not set: read and test it
            ioctl(fd, KDGKBENT, &ke);
            if(ke.kb_value != kmap[i])
            {
                close(fd);
                return -i;
            }
        }

        ke.kb_table=ALTGR_PRESSED;
        if (ioctl(fd,KDSKBENT,&ke))
        {
            //  actioncode not set: read and test it
            ioctl(fd, KDGKBENT, &ke);
            if(ke.kb_value != kmap[i])
            {
                close(fd);
                return -i;
            }
        }

        ke.kb_table=CTRL_PRESSED;
        if (ioctl(fd,KDSKBENT,&ke))
        {
            //  actioncode not set: read and test it
            ioctl(fd, KDGKBENT, &ke);
            if(ke.kb_value != kmap[i])
            {
                close(fd);
                return -i;
            }
        }

        ke.kb_table=SHIFTCTRL_PRESSED;
        if (ioctl(fd,KDSKBENT,&ke))
        {
            //  actioncode not set: read and test it
            ioctl(fd, KDGKBENT, &ke);
            if(ke.kb_value != kmap[i])
            {
                close(fd);
                return -i;
            }
        }

        ke.kb_table=ALT_PRESSED;
        if (ioctl(fd,KDSKBENT,&ke))
        {
            //  actioncode not set: read and test it
            ioctl(fd, KDGKBENT, &ke);
            if(ke.kb_value != kmap[i])
            {
                close(fd);
                return -i;
            }
        }
    }

    //  ubuntu annoyance: altgr-ctrl-fn
    //  does not switch to console n but to n+12

    for (i=59;i<69;i++)
    {
        //  F1, F2,.., F10
        ke.kb_index=i;

        //  read from table 12 (alt-ctrl)
        ke.kb_table=12;

        if (ioctl(fd, KDGKBENT, &ke))
        {
            close(fd);
            printf("\n\t*UNKNOWN READING ERROR* \n");
            return -i;
        }

        //  store actioncode
        kmap[i]=ke.kb_value;

        // write to table  6 (altgr-ctrl)
        ke.kb_table=6;

        if (ioctl(fd,KDSKBENT,&ke))
        {
            //  actioncode not set: read and test it
            ioctl(fd, KDGKBENT, &ke);
            if(ke.kb_value != kmap[i])
            {
                close(fd);
                return -i;
            }
        }
    }

    for (i=87;i<89;i++)
    {
        //  F11, F12
        ke.kb_index=i;

        //  read from table 12 (alt-ctrl)
        ke.kb_table=12;

        if (ioctl(fd, KDGKBENT, &ke))
        {
            close(fd);
            printf("\n\t*UNKNOWN READING ERROR* \n");
            return -i;
        }

        //  store actioncode
        kmap[i]=ke.kb_value;

        // write to table  6 (altgr-ctrl)
        ke.kb_table=6;

        if (ioctl(fd,KDSKBENT,&ke))
        {
            //  actioncode not set: read and test it
            ioctl(fd, KDGKBENT, &ke);
            if(ke.kb_value != kmap[i])
            {
                close(fd);
                return -i;
            }
        }
    }

    //  ubuntu's Caps Lock annoyance, we disable alls Caps Lock combos

    for (i=0; i < 128; i++)
    {
        ke.kb_table = i;
        ke.kb_index = 58;       // caps lock keycode
        ke.kb_value = 0x0200;   // VoidSymbol

        ioctl(fd,KDSKBENT,&ke);
    }

    return 0;
    //  and what is the total effect of above?
    //  read the file kmap_worries.txt
}
